"""
A general tool for plotting, slicing, and analyzing xarray DataArrays.
"""

# ----------------------------------------------------------------------------------

#from iautil.utilities.ui import DataArrayImageView
import numpy as np
import pyqtgraph as pg
from pyqtgraph import dockarea, QtGui, QtCore
import sys
import xarray as xr

# ----------------------------------------------------------------------------------

__all__ = (
    "ImageTool",
    "ImageToolWidget"
)

# ----------------------------------------------------------------------------------

class ImageTool:
    """
    
    """

    app = pg.mkQApp("ImageTool")

    def __init__(self, data_array: xr.DataArray) -> None:
        self.image_tool_widget = ImageToolWidget(data_array)

    def show(self):
        self.image_tool_widget.show()
        self.app.exec_()

# ----------------------------------------------------------------------------------

class ImageToolWidget(QtGui.QWidget):
    """

    """

    def __init__(self, data_array: xr.DataArray) -> None:
        super(ImageToolWidget, self).__init__()

        self.data_array = data_array

        self.data_array_image_view = None
        self.controller_widget = None

        self.dock_area = dockarea.DockArea()
        self.data_array_image_view_dock = None
        self.controller_widget_dock = None
        
        self._create_widgets()
        self._create_docks()

        self.layout = QtGui.QVBoxLayout()
        self.layout.addWidget(self.dock_area)
        self.setLayout(self.layout)

    # ------------------------------------------------------------------------------

    def _create_widgets(self) -> None:
        """
        
        """
        from iautil.plotting.image_tool.controller import DataArrayController
        from iautil.utilities.ui import DataArrayImageView
        
        self.data_array_image_view = DataArrayImageView()
        self.controller_widget = DataArrayController(self.data_array, self)
        
    # ------------------------------------------------------------------------------
    
    def _create_docks(self) -> None:
        """
        
        """

        self.data_array_image_view_dock = dockarea.Dock(
            name="DataArray ImageView",
            size=(200, 200),
            widget=self.data_array_image_view,
            hideTitle=True
        )

        self.controller_widget_dock = dockarea.Dock(
            name="Controller",
            size=(200, 100),
            widget=self.controller_widget,
            hideTitle=True
        )

        self.dock_area.addDock(self.data_array_image_view_dock)
        self.dock_area.addDock(self.controller_widget_dock)

# ----------------------------------------------------------------------------------



"""
Basic I/O functions for creating/loading .iau files and reading from 
various file types.
"""

# ----------------------------------------------------------------------------------

import ast
import h5py
import numpy as np
import os
from typing import List, Tuple
import vtk
from vtk.util import numpy_support as npSup # type: ignore
import xarray as xr

# ----------------------------------------------------------------------------------

__all__ = (
    "create_iau_file",
    "load_iau_file"
)

# ----------------------------------------------------------------------------------

def create_iau_file (
    data_source: str, 
    iau_file_path: str, 
    axis_labels: List[str] = None, 
    new_axis_values: List = None,
    metadata: dict = None
) -> None:
    """
    Creates .iau file from data source. Data source can either be a single file
    or a directory of multiple data source files.

    Parameters:
    data_source (str): File/directory path used to create .iau file from.
    iau_file_path (str): File path to save .iau file in.
    axis_labels (list[str]): Labels for each axis.
    new_axis_values (list): For .iau files created from multiple data source files
        stitched together. Defines values for new axis.
    metadata (dict): Metadata for .iau file.
    """

    # Data source validation
    if not os.path.exists(data_source):
        raise OSError(
            f"{data_source} is an invalid path."
        )

    # Load data and axes from data source
    data, axes = None, None
    
    # Data source as directory
    if os.path.isdir(data_source):
        file_list = os.listdir(data_source) # directory contents, sorted
        file_list.sort()
        data_list, axes_list = [], []

        for file in file_list:
            data, axes = _load_data_source(str(os.path.join(data_source, file)))
            data_list.append(data)
            axes_list.append(axes)
        data, axes = _stitch(data_list, axes_list)

        # Handles new axis values 
        if new_axis_values is None:
            new_axis_values = [i for i in range(data.shape[-1])]
        axes.append(new_axis_values)
    # Data source as file
    elif os.path.isfile(data_source):
        data, axes = _load_data_source(data_source)

    # Handles axis labels
    if axis_labels is None:
        axis_labels = [f"axis_{i}" for i in range(data.ndim)]
    elif len(axis_labels) != data.ndim:
        raise ValueError(
            f"Invalid number of axis labels. Expected # of labels: {data.ndim}"
        )

    # Creates .iau file
    with h5py.File(iau_file_path, 'a') as new_file:
        new_file.create_dataset("data", data=data)
        new_file.attrs["metadata"] = str(metadata)
        new_file.create_group("axes")

        for i in range(len(axes)):
            axis = np.array(axes[i])
            new_file.create_dataset(f"axes/axis_{i}", data=axis)
            new_file["data"].dims[i].label = axis_labels[i]
            new_file[f"axes/axis_{i}"].make_scale(axis_labels[i])
            new_file["data"].dims[i].attach_scale(new_file[f"axes/axis_{i}"])

# ----------------------------------------------------------------------------------

def load_iau_file(file: str) -> xr.DataArray:
    """
    Retrieves data, axis info, and metadata from .iau file in an xarray dataset.

    Parameters:
        file (str): .iau file to load.

    Returns:
        data_array (xr.DataArray): Dataset containing data, axis info, and metadata.
    """
    
    # Reads info from .iau file
    with h5py.File(file, 'r') as iau_file:
        data = iau_file["data"][...]
        axes = [iau_file["data"].dims[i][0][...] for i in range(data.ndim)]
        axis_labels = [iau_file["data"].dims[i].label for i in range(data.ndim)]
        metadata = ast.literal_eval(iau_file.attrs["metadata"])

    # Creates xarray DataArray from .iau info
    # Internal data structure for everything in image-analysis-util
    data_array = xr.DataArray(
        data=data, 
        coords=axes, 
        dims=axis_labels, 
        attrs=metadata
    )

    return data_array
    
# ----------------------------------------------------------------------------------

def _load_data_source(file: str) -> Tuple[np.ndarray, List[list]]:
    """
    Retrieves data and axis values from data source file.

    Parameters:
        file (str): Data source file to load.

    Returns:
        data (np.ndarray): Multi-dimensional NumPy array holding data.
        axes (list[list]): List of values for each axis.
    """
    
    file_ext = os.path.splitext(file)[1]
    data, axes = None, None

    if file_ext == ".vti":
        data, axes = _load_vti(file)

    return data, axes

# ----------------------------------------------------------------------------------

def _stitch(
    data_list: List[np.ndarray],
    axes_list: List[list]
) -> Tuple[np.ndarray, List[list]]:
    """
    _stitches data from multiple data source files. Also checks for inconsistencies 
    in axis values.

    Parameters:
        data_list (list[np.ndarray]): List of data from each data source file.
        axes_list (list[list]): List of axis values from each data source file.

    Returns:
        data (np.ndarray): Stacked data arrays; n + 1 dimensions.
        axes (list[list]): Axis values from each axis.
    """
    
    data, axes = None, None

    # Checks if axes stay consistent throughout data source files
    consistent_axes = axes_list.count(axes_list[0]) == len(axes_list)

    if consistent_axes:
        axes = axes_list[0]
    else:
        raise ValueError(
            "Inconsistent axes throughout data source files."
        )

    # Converts list of NumPy arrays (ndim = n) to one NumPy array (ndim = n + 1)
    data = np.stack(data_list, axis=-1)
    
    return data, axes

# ----------------------------------------------------------------------------------

def _load_vti(file: str) -> Tuple[np.ndarray, List[list]]:
    """
    Retrieves data, axis values from .vti (VTK XML image format) file.
    .vti files contain 2D or 3D datasets.

    Parameters:
        file (str): .vti file to load.

    Returns:
        data (np.ndarray): Multi-dimensional NumPy array holding data.
        axes (list[list]): List of values for each axis.
    """
    
    data_reader = vtk.vtkXMLImageDataReader()
    data_reader.SetFileName(file)
    data_reader.Update()

    raw_data = data_reader.GetOutput()
    dimensions = list(raw_data.GetDimensions())

    data = npSup.vtk_to_numpy(raw_data.GetPointData().GetArray('Scalars_'))
    data = data.reshape(dimensions)

    origin = raw_data.GetOrigin() # First point for each axis
    spacing = raw_data.GetSpacing() # Space between points for each axis
    extent = raw_data.GetExtent() # First and last index of each axis

    axis_0, axis_1, axis_2 = [], [], []

    # Adds values to each axis accordingly
    for point in range(extent[0], extent[1] + 1):
        axis_0.append(origin[0] + point * spacing[0])
    for point in range(extent[2], extent[3] + 1):
        axis_1.append(origin[1] + point * spacing[1])
    for point in range(extent[4], extent[5] + 1):
        axis_2.append(origin[2] + point * spacing[2])

    # A list of lists of varying lengths
    axes = [axis_0, axis_1, axis_2]

    return data, axes

# ----------------------------------------------------------------------------------

"""create_iau_file(
    data_source="./examples/example_files",
    iau_file_path="./examples/example_files/scans.iau",
    axis_labels=["H", "K", "L", "V"],
    metadata={"name": "scans"}
)"""

#load_iau_file("./examples/example_files/scans.iau")



"""
Widget that controls the DataArray loaded into an ImageTool instance.
"""

# ----------------------------------------------------------------------------------

import numpy as np
import pyqtgraph as pg
from PyQt5 import QtGui, QtCore
import xarray as xr

# ----------------------------------------------------------------------------------

__all__ = (
    "DataArrayController"
)

# ----------------------------------------------------------------------------------

AXES = ["x", "y", "z", "t"]

# ----------------------------------------------------------------------------------

class DataArrayController(QtGui.QWidget):
    """
    Controller for the ImageTool DataArray.
    """

    def __init__(self, data_array: xr.DataArray, parent=None) -> None:
        super(DataArrayController, self).__init__(parent)

        self.parent = parent
        self.data_array = data_array
        
        self.layout = DataArrayControllerLayout(data_array, parent=self)
        self.setLayout(self.layout)

        self.dim_lbl_list = self.layout.dim_lbl_list
        self.dim_axis_cbx_list = self.layout.dim_axis_cbx_list
        self.dim_slider_list = self.layout.dim_slider_list
        self.dim_currval_cbx_list = self.layout.dim_currval_cbx_list

        self._update_currval()
        self._update_axes()

    # ------------------------------------------------------------------------------

    def _update_currval(self) -> None:
        """
        
        """
        
        if isinstance(self.sender(), QtGui.QSlider):
            dim = self.dim_slider_list.index(self.sender())
            currval_index = self.sender().value()
            self.dim_currval_cbx_list[dim].setCurrentIndex(currval_index)

        if isinstance(self.sender(), QtGui.QComboBox):
            dim = self.dim_currval_cbx_list.index(self.sender())
            currval_index = self.sender().currentIndex()
            self.dim_slider_list[dim].setValue(currval_index)

        self._update_image_view()

    # ------------------------------------------------------------------------------

    def _update_axes(self) -> None:
        """
        
        """

        if not self.sender() is None:
            new_axis = self.sender().currentIndex()
            axes = [i for i in range(self.data_array.ndim)]
            curr_axes = [cbx.currentIndex() for cbx in self.dim_axis_cbx_list]
            
            try:
                axis_to_add = list(set(axes) - set(curr_axes))[0]
            except:
                pass

        for i in range(self.data_array.ndim):
            cbx = self.dim_axis_cbx_list[i]

            if not self.sender() is None:
                if cbx.currentIndex() == new_axis and not cbx == self.sender():
                    cbx.setCurrentIndex(axis_to_add)

            if cbx.currentIndex() <= 1:
                self.dim_slider_list[i].setEnabled(False)
                self.dim_currval_cbx_list[i].setEnabled(False)
            else:
                self.dim_slider_list[i].setEnabled(True)
                self.dim_currval_cbx_list[i].setEnabled(True)

        self._update_image_view()

    # ------------------------------------------------------------------------------

    def _update_image_view(self) -> None:
        """
        
        """

        image_view_image = None
        str_numpy_args = ""
        transpose = False
        x_index, y_index = 0, 0

        for i in range(self.data_array.ndim):
            if self.dim_slider_list[i].isEnabled():
                str_numpy_args += f"{self.dim_slider_list[i].value()}"
            else:
                str_numpy_args += ":"
                if self.dim_axis_cbx_list[i].currentIndex() == 0:
                    x_index = i
                else:
                    y_index = i

            if i < self.data_array.ndim - 1:
                str_numpy_args += ","

        if y_index < x_index:
            transpose = True
        
        numpy_args = eval(f'np.s_[{str_numpy_args}]')
        data_array_slice = self.data_array[numpy_args]

        if data_array_slice.values.ndim != 0:
            if not transpose:
                self.parent.data_array_image_view.set_data_array(data_array_slice)
            else:
                self.parent.data_array_image_view.set_data_array(data_array_slice.T)
        
# ----------------------------------------------------------------------------------

class DataArrayControllerLayout(QtGui.QGridLayout):
    """
    
    """

    def __init__(self, data_array: xr.DataArray, parent=None) -> None:
        super(DataArrayControllerLayout, self).__init__(parent)

        self.dim_lbl_list = []
        self.dim_axis_cbx_list = []
        self.dim_slider_list = []
        self.dim_currval_cbx_list = []
        axes = AXES[:data_array.ndim]

        for i in range(data_array.ndim):
            dim_lbl = data_array.dims[i]
            dim_coords = map(str, data_array.coords[dim_lbl].values)

            self.dim_lbl_list.append(QtGui.QLabel(dim_lbl))
            self.dim_axis_cbx_list.append(QtGui.QComboBox())
            self.dim_slider_list.append(QtGui.QSlider(QtCore.Qt.Horizontal))
            self.dim_currval_cbx_list.append(QtGui.QComboBox())

            self.dim_axis_cbx_list[i].addItems(axes)
            self.dim_axis_cbx_list[i].setCurrentIndex(i)
            self.dim_slider_list[i].setMaximum(data_array.shape[i] - 1)
            self.dim_currval_cbx_list[i].addItems(dim_coords)

            self.addWidget(self.dim_lbl_list[i], i, 0)
            self.addWidget(self.dim_axis_cbx_list[i], i, 1)
            self.addWidget(self.dim_slider_list[i], i, 2, 1, 3)
            self.addWidget(self.dim_currval_cbx_list[i], i, 5)

            self.dim_axis_cbx_list[i].currentIndexChanged.connect(
                parent._update_axes
            )
            self.dim_slider_list[i].valueChanged.connect(
                parent._update_currval
            )
            self.dim_currval_cbx_list[i].currentIndexChanged.connect(
                parent._update_currval
            )

# ----------------------------------------------------------------------------------


"""

"""

# ----------------------------------------------------------------------------------

from setuptools import setup, find_packages

# ----------------------------------------------------------------------------------

setup( 
    name="image-analysis-util",
    version="0.0.18",
    description="Scientific image exploration software for multidimensional datasets.",
    author="Henry Smith",
    author_email="smithh@anl.gov",
    url="https://github.com/henryjsmith12/image-analysis-util/iautil",
    install_requires=[
        "h5py",
        "numpy",
        "PyQt5",
        "pyqtgraph",
        "scipy",
        "vtk",
        "xarray",
    ],
    packages=find_packages(),
    license="See LICENSE file",
    platforms="any",
)




"""
UI widget classes.
"""

# ----------------------------------------------------------------------------------

from matplotlib import colors
from matplotlib import pyplot as plt
import numpy as np
import pyqtgraph as pg
from pyqtgraph import QtGui, QtCore
import xarray as xr

# ----------------------------------------------------------------------------------

__all__ = (
    "DataArrayImageView",
    "DataArrayPlot",
    "set_data_array"
)

# ----------------------------------------------------------------------------------

class DataArrayImageView(pg.ImageView):
    """
    A custom PyQtGraph ImageView.
    """
    
    def __init__(self, parent=None) -> None:
        super(DataArrayImageView, self).__init__(
            parent, 
            view=pg.PlotItem()
        )

        self.ui.histogram.hide()
        self.ui.roiBtn.hide()
        self.ui.menuBtn.hide()
        
        self.view.setAspectLocked(lock=False)
        self.view.enableAutoRange()

    # ------------------------------------------------------------------------------

    def set_data_array(self, data_array: xr.DataArray) -> None:
        """
        
        """

        image = self._set_color_map(data_array.values)
        self.view.setLabels(
            bottom = data_array.dims[0],
            left = data_array.dims[1]
        )

        self.setImage(image)

    # ------------------------------------------------------------------------------

    def _set_color_map(image: np.ndarray) -> np.ndarray:
        """
        
        """

        image_max = np.amax(image)
        norm = colors.LogNorm(vmax=image_max)

        normalized_image = norm(image)
        color_image = plt.cm.jet(normalized_image)

        return color_image

# ----------------------------------------------------------------------------------

class DataArrayPlot(pg.PlotWidget):
    """
    A custom PyQtGraph PlotWidget.
    """
    
    def __init__(self, parent=None, plotItem=None) -> None:
        super(DataArrayPlot, self).__init__(parent, plotItem)

# ----------------------------------------------------------------------------------